/*************************************************************************
 Decompresses compressed Nintendo DS Download Station demos. 
 Written by Michael Chisholm ("Chishm")
 Distributed under the GNU GPL

 Based on testmini.c, 
 copyright (C) 2005 Markus Franz Xaver Johannes Oberhumer.
**************************************************************************/

#include "minilzo.h"
#include <string.h>
#include <stdio.h>
#include <stdlib.h>

const unsigned char lzo_header[] = {'L', 'Z' , 'O', 'n', 0x00, 0x2F, 0xF1, 0x71};

void usage(char *prog) {
	fprintf(stderr, "  Usage: %s <in>\n\n", prog);
}

int main(int argc, char *argv[])
{
	FILE *input_file;
	FILE *output_file;
    int r;
    lzo_uint in_len;
    lzo_uint out_len;
    lzo_uint new_len;

	unsigned char lzo_compressed_size[4];
	unsigned char lzo_decompressed_size[4];

	unsigned char lzo_file_header[8];

	unsigned char *in;
	unsigned char *out;

	char* outfile_name;
	size_t outfile_name_length;

	printf ("Nintendo DS Download Station demo decompressor\n");
	printf ("Written by Michael Chisholm (\"Chishm\")\n");
	printf ("Distributed under the GNU GPL\n");

	if (argc != 2) {
		usage(argv[0]);
		return -1;
	}

	printf("\nUses LZO real-time data compression library (v%s, %s).\n",
            lzo_version_string(), lzo_version_date());
    printf("Copyright (C) 1996-2005 Markus Franz Xaver Johannes Oberhumer\nAll Rights Reserved.\n\n");


	// Init the lzo library
	if (lzo_init() != LZO_E_OK)
    {
        fprintf(stderr, "internal error - lzo_init() failed !!!\n");
        return 3;
    }

	if ((input_file = fopen (argv[1], "rb")) == NULL) {
		fprintf(stderr, "  ERROR: Could not open input file (%s)!\n", argv[1]);
		return -1;
	}

	fseek (input_file, 0x00, SEEK_SET);
	fread (lzo_file_header, sizeof(unsigned char), 8, input_file);

	if (memcmp (lzo_header, lzo_file_header, 8) != 0) {
		fprintf(stderr, "  ERROR: Not a valid compressed demo (%s)!\n", argv[1]);
		return -1;
	}

	// Get input size
	fseek (input_file, 0x0C, SEEK_SET);
	fread (lzo_compressed_size, sizeof(unsigned char), 4, input_file);
	in_len = (lzo_compressed_size[3]) + (lzo_compressed_size[2] << 8) + (lzo_compressed_size[1] << 16) + (lzo_compressed_size[0] << 24);
	printf("In size: %x\n", in_len);

	// Get output size
	fseek (input_file, 0x08, SEEK_SET);
	fread (lzo_decompressed_size, sizeof(unsigned char), 4, input_file);
	out_len = (lzo_decompressed_size[3]) + (lzo_decompressed_size[2] << 8) + (lzo_decompressed_size[1] << 16) + (lzo_decompressed_size[0] << 24);
	printf("Out size: %x\n", out_len);

	// Read in compressed data
	in = malloc (in_len);
	fseek (input_file, 0x10, SEEK_SET);
	fread (in, sizeof(unsigned char), in_len, input_file);

	// Allocate space for decompressed data
	out = malloc (out_len);
	memset (out, 0, out_len);

	// Finished with input file
	fclose (input_file);


	// Decompress the demo
    r = lzo1x_decompress(in,in_len,out,&new_len,NULL);
	printf("Actual size: %x\n", new_len);
    if (r == LZO_E_OK && new_len <= out_len)
        printf("Decompressed %lu bytes back into %lu bytes\n",
            (unsigned long) in_len, (unsigned long) out_len);
    else
    {
        /* this should NEVER happen */
        fprintf(stderr, "internal error - decompression failed: %d\n", r);
        return 1;
    }

	free (in);

	// Write decompressed data to a file
	outfile_name_length = strlen(argv[1]) + strlen(".nds");
	outfile_name = malloc(outfile_name_length);
	strncpy(outfile_name, argv[1], outfile_name_length);
	strncat(outfile_name, ".nds", outfile_name_length);

	if ((output_file = fopen(outfile_name, "wb")) == NULL) {
		fprintf(stderr, "  ERROR: Could not open output file (%s)!\n", argv[2]);
		return -1;
	}

	fwrite (out, sizeof(unsigned char), out_len, output_file);
	fclose (output_file);

	free (out);

	printf("Decompressed data written to %s\n", outfile_name);
	free (outfile_name);

    return 0;
}

